/*
  Copyright (c) 2015, Matthias Schiffer <mschiffer@universe-factory.net>
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.
    2. Redistributions in binary form must reproduce the above copyright notice,
       this list of conditions and the following disclaimer in the documentation
       and/or other materials provided with the distribution.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/


#pragma once

#include "grammar.hpp"

#include <stdexcept>
#include <set>
#include <map>


namespace solar {

class generator_t {
public:
	class conflict_error : public std::runtime_error {
	public:
		explicit conflict_error(const std::string& what_arg) : std::runtime_error(what_arg) {}
		explicit conflict_error(const char* what_arg) : std::runtime_error(what_arg) {}
	};

private:
	grammar_t grammar;

	std::map<item_t, size_t> rule_ids;
	std::multimap<std::string, size_t> nonterms;

	std::set<std::string> nonterminals;
	std::set<symbol_t> terminals;

	std::multimap<symbol_t, item_t> items;
	std::map<std::set<item_t>, size_t> itemsets;

	std::map<std::pair<size_t, symbol_t>, size_t> shifts;
	std::map<std::pair<size_t, std::string>, size_t> gotos;


	void close_set(std::set<item_t> *set);
	std::set<item_t> get_set(const std::string &nonterm);

	std::pair<std::map<std::set<item_t>, size_t>::iterator, bool> add_set(const std::set<item_t> &set) {
		return itemsets.insert(std::make_pair(set, itemsets.size()));
	}

	void add_shift(size_t from, const symbol_t &sym, size_t to) {
		if (has_reduce_conflict(from, sym))
			throw conflict_error("shift/reduce conflict");

		shifts.insert(std::make_pair(std::make_pair(from, sym), to));
	}

	void add_goto(size_t from, const std::string &nonterm, size_t to) {
		gotos.insert(std::make_pair(std::make_pair(from, nonterm), to));
	}

protected:
	virtual bool has_reduce_conflict(size_t from, const symbol_t &sym) = 0;
	virtual void add_reduction(size_t from, size_t rule) = 0;

	generator_t(const grammar_t &grammar0);

public:
	const grammar_t & get_grammar() const {
		return grammar;
	}

	const std::set<std::string> & get_nonterminals() const {
		return nonterminals;
	}

	const std::set<symbol_t> & get_terminals() const {
		return terminals;
	}

	size_t get_state_count() const {
		return itemsets.size();
	}

	const std::map<std::pair<size_t, symbol_t>, size_t> & get_shifts() const {
		return shifts;
	}

	const std::map<std::pair<size_t, std::string>, size_t> & get_gotos() const {
		return gotos;
	}

	virtual ~generator_t() {}

	void generate();
};

}
